# Copyright 2013 Johan Rydberg.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
import os

from fabric.api import sudo, settings, hide
from fabric.network import disconnect_all


log = logging.getLogger(__name__)


class Configure(object):

    def __init__(self, username, ssh_key_file):
        self.username = username
        self.ssh_key_file = ssh_key_file

    def docker_run(self, image, command=None, ports=None, binds=None, env=None,
                   detach=True, open_stdin=False, tty=False):
        """Run a docker container."""
        options = []
        if detach:
            options.append('-d')
        if open_stdin:
            options.append('-i')
        if tty:
            options.append('-t')
        if ports:
            for port in ports:
                options.extend(['-p', port])
        if binds:
            for bind in binds:
                options.extend(['-v', bind])
        if env:
            for var, val in env.items():
                options.extend(['-e', '"{0}={1}"'.format(var, val)])
        sudo('docker -H 127.0.0.1:3000 run {options} {image} {command}'.format(
                options=' '.join(options), image=image,
                command=command or ''))
        
    @contextlib.contextmanager
    def configure(self, host):
        key_filename = os.path.expanduser(self.ssh_key_file)
        try:
            with settings(key_filename=key_filename,
                          user=self.username,
                          host_string=host):
                self._init()
                yield
        finally:
            with hide('status'):
                disconnect_all()

    @contextlib.contextmanager
    def enter(self, host):
        key_filename = os.path.expanduser(self.ssh_key_file)
        try:
            with settings(key_filename=key_filename,
                          user=self.username,
                          host_string=host):
                yield
        finally:
            with hide('status'):
                disconnect_all()

    def _init(self):
        """Perform basic initialization of the host; installs and
        starts docker.
        """
        sudo('curl https://get.docker.io/gpg | apt-key add -')
        sudo('echo "deb http://get.docker.io/ubuntu docker main" > /etc/apt/sources.list.d/docker.list')
        sudo('apt-get -qq update ')
        sudo('apt-get -qq install -y linux-image-extra-$(uname -r)')
        sudo('apt-get install -y lxc-docker')
        # XXX: right not we're running over HTTP to support WebSocket.
        sudo('sed -i "s#docker -d#docker -d -H 0.0.0.0:3000#g" /etc/init/docker.conf')
        sudo('service docker restart')
        sudo('modprobe aufs')



# Copyright 2013 Johan Rydberg.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

import versioneer
versioneer.versionfile_source = "gilliam_aws/_version.py"
versioneer.versionfile_build = "gilliam_aws/_version.py"
versioneer.tag_prefix = ""
versioneer.parentdir_prefix = ""
commands = versioneer.get_cmdclass().copy()

setup(
    name="gilliam-aws",
    version=versioneer.get_version(),
    cmdclass=commands,
    packages=find_packages(),
    author="Johan Rydberg",
    author_email="johan.rydberg@gmail.com",
    description="Command-line tool for running Gilliam on AWS",
    license="Apache 2.0",
    keywords="app platform",
    url="https://github.com/gilliam/",
    install_requires=['boto', 'fabric', 'requests', 'gilliam-py', 'gilliam-cli'],
    entry_points={
        'gilliam.commands': [
            'aws create = gilliam_aws.commands:Create',
            'aws status = gilliam_aws.commands:Status',
            'aws destroy = gilliam_aws.commands:Destroy',
            ]
        },
)



# Copyright 2013 Johan Rydberg.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import os

from boto.ec2 import connect_to_region


log = logging.getLogger(__name__)


AMI_MAPPING = {
    #'ap-northeast-1': 'ami-6b01666a',
    #'ap-southeast-1': 'ami-1cdf8a4e',
    'eu-west-1': 'ami-2adc3c5d',
    #'sa-east-1': 'ami-ab3791b6',
    #'us-east-1': 'ami-69d9fc00',
    ##'us-west-1': 'ami-40aa9c05',
    #'us-west-2': 'ami-b8b92288',
    }


def connect(region, **args):
    """Create a EC2 connection to a specific region.

    :returns: The EC2 connection object
    """
    return connect_to_region(region, **args)


def _get_or_make_group(conn, name):
    groups = conn.get_all_security_groups()
    group = [g for g in groups if g.name == name]
    if len(group) > 0:
        return group[0]
    else:
        log.info("creating security group %s" % (name,))
        return conn.create_security_group(name, "Gilliam EC2 group")


def _create_security_groups(conn, prefix, allowed, spec):
    groups = {name: _get_or_make_group(
                        conn, '{0}-{1}'.format(prefix, name))
              for name in spec.keys()}

    for name, rules in spec.items():
        group = groups.get(name)
        if group.rules:
            continue
        for rule in rules:
            if type(rule) in (tuple, list):
                for allow in allowed:
                    rule = list(rule) + [allow]
                    group.authorize(*rule)
            else:
                group.authorize(src_group=groups.get(rule))
    return groups


def _get_or_make_keypair(conn, key_dir, key_name):
    """Get or create a key pair with the given name."""
    try:
        key = conn.get_all_key_pairs(keynames=[key_name])[0]
    except conn.ResponseError, e:
        if e.code == 'InvalidKeyPair.NotFound':
            log.info("creating keypair {0}".format(key_name))
            # Create an SSH key to use when logging into instances.
            key = conn.create_key_pair(key_name)

            # Make sure the specified key_dir actually exists.
            # If not, create it.
            key_dir = os.path.expanduser(key_dir)
            key_dir = os.path.expandvars(key_dir)
            if not os.path.isdir(key_dir):
                os.makedirs(key_dir, 0700)

            # AWS will store the public key but the private key is
            # generated and returned and needs to be stored locally.
            # The save method will also chmod the file to protect
            # your private key.
            key.save(key_dir)
        else:
            raise
    return key


def _wait_for_system_and_instance_status_checks(conn, instances):
    """Wait for the given instances to pass system and instance status
    checks.
    """
    log.info("waiting for instances to pass system and status checks...")

    instance_ids = [i.id for i in instances]
    while True:
        statuses = conn.get_all_instance_status(instance_ids)
        pending = [status for status in statuses
                   if (status.system_status.status != 'ok'
                       or status.instance_status.status != 'ok')]
        if not pending:
            break
        time.sleep(5)


def _wait_for_instances_to_become_running(conn, instances):
    """Wait for given instances to become running."""
    while True:
        for i in instances:
            i.update()
        if len([i for i in instances if i.state == 'pending']) > 0:
            time.sleep(5)
        else:
            break

def _wait_for_instances(conn, instances):
    """Wait for instances to become fully ready."""
    _wait_for_instances_to_become_running(conn, instances)
    _wait_for_system_and_instance_status_checks(conn, instances)


def _reserve_instances(conn, config, security_groups, key_name):
    """Create instances based on the given configuration.  This
    implementation creates a single instance that has provides every
    role.

    :returns: The :class:`boto.ec2.instance.Reservation`.
    """
    ami = AMI_MAPPING[config.get('aws_region')]
    image = conn.get_all_images(image_ids=[ami])[0]
    return image.run(
        key_name=key_name,
        security_groups=security_groups.values(),
        instance_type=config.get('aws_ec2_instance_type'),
        min_count=1,
        max_count=1)


def _collect_instances(conn, name):
    """Givn a EC2 connection and a name, collect instances that
    belong to that stage.  Will only collect active instances.
    """
    instances = []
    for reservation in conn.get_all_instances():
        group_names = [g.name for g in reservation.groups]
        if any([group_name.startswith(name + '-') for group_name in group_names]):
            instances.extend(
                i for i in reservation.instances)
                #if is_active(i))
    return instances


# Check whether a given EC2 instance object is in a state we consider active,
# i.e. not terminating or terminated. We count both stopping and stopped as
# active since we can restart stopped clusters.
def is_active(instance):
    return (instance.state in ['pending', 'running', 'stopping', 'stopped'])


class AmazonWebServicesStage(object):

    SECURITY_GROUPS = {
        'router': [
            ('tcp', 22, 22),
            ('tcp', 8080, 8080),
            ],
        'exec': [
            'exec', 'router',
            ('tcp', 22, 22),
            ('tcp', 9000, 9000),
            ('tcp', 49153, 65535),    # the complete Docker port range
            ],
        'sr': [
            'exec', 'router',
            ('tcp', 3222, 3222)
            ],
        }


    def __init__(self, config, name, nodes, ssh_key_file=None):
        self.config = config
        self.name = name
        self.nodes = nodes
        self.username = 'ubuntu'
        self.ssh_key_file = ssh_key_file

    @classmethod
    def get(cls, conn, config, name):
        """Get an existing cluster if available."""
        nodes = _collect_instances(conn, name)
        if nodes:
            return cls(config, name, nodes)
        else:
            return None

    @classmethod
    def create(cls, conn, config, name, allowed=['0.0.0.0/0']):
        """
        Create a new stage running on Amazon Web Services. The stage
        config `config` provides data needed to bootstrap the stage.

        :param conn: AWS EC2 connection.
        :type conn: :class:`boto.ec2.connection.EC2Connection`.

        :param config: Stage configuration.
        :type config: :class:`gilliam_client.config.StageConfig`.

        :param name: The name of the stage.
        :type name: `str`.

        :returns: the created `AmazonWebServicesStage` object.
        """
        log.info("creating stage {0}".format(name))
        # FIXME: the path should not be specified here.
        key_name = name
        key_dir = os.path.expanduser("~/.gilliam/ec2-ssh-keys")
        key_pair = _get_or_make_keypair(conn, key_dir, key_name)
        security_groups = _create_security_groups(
            conn, name, allowed, AmazonWebServicesStage.SECURITY_GROUPS)
        res = _reserve_instances(conn, config, security_groups, key_name)
        _wait_for_instances(conn, res.instances)
        return cls(config, name, res.instances, ssh_key_file=os.path.join(
                key_dir, name + '.pem'))

    def destroy(self, conn):
        """Destroy the cluster by terminating all instances."""
        for inst in self.nodes:
            if inst.state not in ["shutting-down", "terminated"]:
                inst.terminate()

    def _roles(self, node):
        """From a EC2 instance try to decuce what roles it has.

        :type node: a :class:`boto.ec2.instance.Instance`
        """
        group_role_map = {
            'sr': 'service-registry',
            'exec': 'executor'
            }

        group_names = [g.name for g in node.groups]
        roles = []
        for group in group_names:
            if not group.startswith(self.name + '-'):
                continue
            role = group[len(self.name) + 1:]
            role = group_role_map.get(role, role)
            roles.append(role)
        return roles

    def iter_roles(self):
        """Return a sequence of `(hostname, roles)` tuples."""
        for node in self.nodes:
            yield node.public_dns_name, self._roles(node)



# Copyright 2013 Johan Rydberg.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import random
import sys

from gilliam_cli.command import Command, ListerCommand
from gilliam_cli.config import StageConfig

from .configure import Configure
from .ec2 import AmazonWebServicesStage, connect


log = logging.getLogger(__name__)


_SERVICE_REGISTRY_IMAGE = 'gilliam/service-registry'
_EXECUTOR_IMAGE = 'gilliam/executor'
_BOOTSTRAP_IMAGE = 'gilliam/bootstrap'
_PROXY_IMAGE = 'gilliam/proxy'


#: Tag of bootstrap image to run.
_DEFAULT_BOOTSTRAP_TAG = 'latest'


def _connect(stage_config):
    return connect(
        stage_config.get('aws_region'),
        aws_access_key_id=stage_config.get('aws_access_key_id'),
        aws_secret_access_key=stage_config.get('aws_secret_access_key'))


class Status(ListerCommand):
    """display status about stage"""

    FIELDS = ('id', 'host', 'state', 'roles', 'launched_at', 'az')

    requires = {'stage': True}
    
    def take_action(self, options):
        conn = _connect(self.app.config.stage_config)

        stage = AmazonWebServicesStage.get(
            conn, self.app.config.stage_config,
            self.app.config.stage
            )

        def it(stage):
            for node in stage.nodes:
                yield (
                    node.id,
                    node.public_dns_name,
                    node.state,
                    ' '.join(stage._roles(node)),
                    node.launch_time,
                    node.placement
                    )

        return self.FIELDS, it(stage)


class Destroy(Command):
    """destroy stage running on AWS"""

    def take_action(self, options):
        conn = _connect(self.app.config.stage_config)
        stage = AmazonWebServicesStage.get(
            conn, self.app.config.stage_config,
            self.app.config.stage
            )
        stage.destroy(conn)


class Create(Command):
    """create a new Gilliam stage running on Amazon Web Services:

      gilliam aws create [options] app-prod

    Credentials for AWS is passed through the `--access-key-id` and
    `--secret-key` options or via the standard environment variables
    (`AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY`).

    For configuration, provide EC2 key pair with `--key-pair` and path
    to the corresponding SSH key file with `--ssh-key-file`.

    The stage is created in the EU east region unless region is
    specified with `--region`.
    """

    def get_parser(self, prog_name):
        parser = Command.get_parser(self, prog_name)
        parser.add_argument('name')
        parser.add_argument('--access-key-id', metavar="DATA")
        parser.add_argument('--secret-access-key', metavar="DATA")
        parser.add_argument('--region', default='us-east-1', metavar="REGION")
        parser.add_argument('--instance-type', default='m1.small', metavar="TYPE")
        parser.add_argument('--repository', metavar="NAME")
        parser.add_argument('-B', '--bootstrap-tag', metavar="TAG",
                            default=_DEFAULT_BOOTSTRAP_TAG)
        return parser

    def take_action(self, options):
        self._check_existing(self.app.config, options)
        stage_config = StageConfig.create(options.name)
        self._check_credentials(stage_config, options)
        self._build_config(stage_config, options)

        # step 1. create resources
        conn = _connect(stage_config)
        stage = AmazonWebServicesStage.create(conn, stage_config,
                                              options.name)

        # step 2. configure resources
        configure = Configure(stage.username, stage.ssh_key_file)
        self._configure(stage, configure)
        self._bootstrap(stage, configure, options.bootstrap_tag)

        # step 3. update stage config
        stage_config.set('service_registry', [
                'http://{0}:3222'.format(hostname)
                for (hostname, roles) in stage.iter_roles()
                if 'service-registry' in roles])
        if options.repository:
            stage_config.set('repository', options.repository)

        # stage 4. profit.
        stage_config.write()

    def _check_existing(self, config, options):
        """Make sure that there isn't a stage with this name already.
        If there is, panic.
        """
        try:
            log.debug("checking for existing stage ...")
            StageConfig.make(options.name)
        except EnvironmentError:
            pass
        else:
            sys.exit("there seem to be a stage with that name already")

    def _check_credentials(self, stage_config, options):
        """Make sure credentials are OK."""
        if not options.access_key_id:
            options.access_key_id = os.getenv('AWS_ACCESS_KEY_ID')
        if not options.secret_access_key:
            options.secret_access_key = os.getenv('AWS_SECRET_ACCESS_KEY')

    def _build_config(self, stage_config, options):
        vars = [('aws_access_key_id', options.access_key_id, True),
                ('aws_secret_access_key', options.secret_access_key, True),
                ('aws_region', options.region, True),
                ('aws_ec2_instance_type', options.instance_type, True)]
        for (var, value, required) in vars:
            if required and not value:
                sys.exit("config var %s is required" % (var,))
            stage_config.set(var, value)

    def _executor_name(self, node):
        return node.split('.')[0]

    def _configure(self, stage, configure):
        """Set up basic configuration such as installing Docker (done
        by `configure`) and install components that lives outside of
        Gilliam such as the service-registry (executor depends on it)
        and the executor (the rest of the system depends on it).
        """
        for hostname, roles in stage.iter_roles():
            log.debug('configuring {0}'.format(hostname))
            with configure.configure(hostname):
                if 'service-registry' in roles:
                    self._start_service_registry(stage, hostname, configure)
                if 'executor' in roles:
                    self._start_proxy(stage, hostname, configure)
                    self._start_executor(stage, hostname, configure)

    def _start_service_registry(self, stage, host, configure):
        service_registry_cluster = self._make_service_registry_option(stage)
        log.debug('launching service registry')
        options = '-n {0} -c {1}'.format(host, service_registry_cluster)
        configure.docker_run(_SERVICE_REGISTRY_IMAGE, options, ports=['3222:3222'])

    def _start_executor(self, stage, host, configure):
        log.debug('launching executor')
        service_registry = self._make_service_registry_option(stage)
        options = '--host {0} --name {1}'.format(host, self._executor_name(host))
        env = {
            'GILLIAM_SERVICE_REGISTRY': service_registry,
            'DOCKER': 'http://{0}:3000'.format(host)
            }
        configure.docker_run(_EXECUTOR_IMAGE, options, env=env, ports=['9000:9000'])

    def _start_proxy(self, stage, host, configure):
        log.debug('launching proxy')
        service_registry = self._make_service_registry_option(stage)
        env = {
            'GILLIAM_SERVICE_REGISTRY': service_registry,
            }
        configure.docker_run(_PROXY_IMAGE, 'bin/proxy', env=env, ports=['9001:9001'])

    def _bootstrap(self, stage, configure, tag):
        """Run bootstrap script that will bring the system to life."""
        env = {
            # The bootstrap script need to know how to talk to the
            # service registry; fill in the service registry
            # environment variable so gilliam-cli knows where to pick
            # up the information.
            'GILLIAM_SERVICE_REGISTRY': self._make_service_registry_option(
                stage),

            # Routers need special attention since they are pinned to
            # specific executors.  The ROUTERS variable will hold a
            # space separated list of executor instance names that
            # should get a dedicated router.
            'ROUTERS': ' '.join(self._executor_name(host)
                                for (host, roles) in stage.iter_roles()
                                if 'router' in roles),
            }

        hostname = random.choice([h for (h, roles) in stage.iter_roles()])
        image = '{0}:{1}'.format(_BOOTSTRAP_IMAGE, tag)
        with configure.enter(hostname):
            log.debug("bootstrapping from {0} using {1}".format(hostname, image))
            configure.docker_run(image, '', env=env, detach=False)

    def _make_service_registry_option(self, stage):
        return ','.join(
            '{0}:3222'.format(hostname)
            for (hostname, roles) in stage.iter_roles()
            if 'service-registry' in roles)
